#ifndef ECOLIST__H
#define ECOLIST__H

#include "sequence.h"
#include <list>

#define MAXSUBNODES

typedef union block_t block;

union block_t {
  word w;
  block_t* p;
};

class Ecolist {
 public:
  typedef pair<int,int> val_type;
  typedef class BidirectionalEcolistIterator {
  public:
    typedef BidirectionalEcolistIterator _Self;

    block *current;
    list<block*> prev;
    int pos,ite,end;
    BidirectionalEcolistIterator(block *start,list<block*> prev,int p,int ite);
    void put(int seq,int idx) const;
    void put(const val_type &v) const;
    val_type operator*() const;
    val_type get() const; //identical to operator*, just there for symmetry with put
    _Self& operator++();
    _Self operator++(int);
    _Self& operator--();
    _Self operator--(int);
    bool operator!=(const _Self&) const;
    bool operator==(const _Self&) const;
    inline int to_i() const {int ret=pos;for(int j=ite;j>0;j--)ret+=minChunks<<(j-1);return ret;}
  } bi_iterator;

  typedef class EcolistIterator {
  public:
    typedef EcolistIterator _Self;

    block *current;
    int pos,ite,end;
    EcolistIterator(block *start,int p,int ite);
    EcolistIterator(const BidirectionalEcolistIterator&);
    void put(int seq,int idx) const;
    void put(const val_type &v) const;
    val_type operator*() const;
    val_type get() const; //identical to operator*, just there for symmetry with put
    _Self& operator++();
    _Self operator++(int);
    bool operator!=(const _Self&) const;
    bool operator==(const _Self&) const;
    inline int to_i() const {int ret=pos;for(int j=ite;j>0;j--)ret+=minChunks<<(j-1);return ret;}
  } iterator;


  //blocks are allocated in blobs (of some 2^n*initBlocks)
  //blocks are composed of chunks.
  //chunks might straddle word/block boundaries.
  static int minBlocks,minChunks;
  static unsigned seqbits,idxbits,chunkSize;
  static word idxLongest;
  static int initBlocks;
  static bool inited;
  static bool (*lessthan)(const pair<int,int>&,const pair<int,int>&);
  static int externalSortLimit;
  
  int size;
  block *blocks;

  //inity stuff
  Ecolist();
  ~Ecolist();
  static void init(int seqs,word longest,word total,int weight,bool (*_lt)(const val_type&,const val_type&));

  //accessy-stuff
  pair<int,int> getValAt(int idx);
  iterator begin();
  iterator end();
  bi_iterator bi_begin();
  bi_iterator bi_end();
  int capacity();
  inline int getSize(){return size;}
  iterator at(int pos);
  bi_iterator bi_at(int pos);

  //take care only to use on sorted lists...
  inline iterator lower_bound(const val_type& m){return at(int_lower_bound(m));}
  inline iterator upper_bound(const val_type& m){return at(int_upper_bound(m));}
  inline bi_iterator bi_lower_bound(const val_type& m){return bi_at(int_lower_bound(m));}
  inline bi_iterator bi_upper_bound(const val_type& m){return bi_at(int_upper_bound(m));}
  int int_lower_bound(const val_type& m); //first elem not <m
  int int_upper_bound(const val_type& m); //first elem x where m<x (or end())
  inline pair<iterator,iterator> equal_range(const val_type& m){return pair<iterator,iterator>(lower_bound(m),upper_bound(m));}

  //manipulators
  void clear(); //erase and reset
  void putValAt(int idx,int seq,int pos);
  void putValAt(int idx,const val_type&);
  void push(int seq,int pos);
  void swap(iterator &a,iterator &b);
  inline void sort(){if(size>externalSortLimit)inPlaceSort(); else externalSort();}
  void inPlaceSort(); //in place sort. might be necessary if size becomes crazy-big
  void externalSort(); //this is faster, but takes more memory

  //behind the scenes mojo
 protected:
  static void eraseAll(block* root,int ite);
  static block* findBlockFor(block* root,int &remain);
  static block* allocBlob(int ite);
  static void writeBits(block* root,int loc,int offset,int size,int val);
  static int readBits(block* root,int loc,int offset,int size);

  int parition(int left,int right,int pivoti);
  void qsort(int left,int right);

  friend ostream& operator<<(ostream& os,Ecolist& a);
};

#endif

